#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils

# the users class.
class Cache(object):
	def __init__(self):	

		# variables.
		self.__cache__ = {}
		
		#
	def set(self, 
		# the cache mode (example: subscriptions) (optional).
		mode=None, 
		# the cache key (example: someuser@gmail.com) (required).
		key=None, 
		# the cached value.
		value=None,
		# the minutes before reset (leave None to use no refresh).
		reset=60*24,
	):
		data = {
			"value":value,
			"reset":reset,
			"timestamp":Date().timestamp,
		}
		if mode != None:
			try:self._cache_[mode]
			except KeyError: self.__cache__[mode] = {}
			self.__cache__[mode][key] = data
		else:
			self.__cache__[key] = data
		return r3sponse.success(f"Successfully cached [{mode}.{key}].") 
	def get(self, 
		# the cache mode (example: subscriptions) (optional).
		mode=None, 
		# the value key (example: someuser@gmail.com) (required).
		key=None, 
	):
		if mode != None:
			try:self.__cache__[mode]
			except KeyError:  return r3sponse.error(f"No cached {mode} data available.") 
			try:data = self.__cache__[mode][key]
			except KeyError:  return r3sponse.error(f"No cached {mode}.{key} data available.") 
		else:
			try:data = self.__cache__[key]
			except KeyError:  return r3sponse.error(f"No cached {key} data available.") 
		if isinstance(data["reset"], int):
			date = Date()
			increased = date.increase(data["timestamp"], minutes=data["reset"])
			if date.compare(increased, date.timestamp) in ["past", "present"]:
				return r3sponse.error(f"Cache refresh required for {mode}.{key}.") 
		return r3sponse.success(f"Successfully retrieved cache [{mode}.{key}].", {
			"value":data["value"],
		}) 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
try:
	from w3bsite.classes.firebase import *
	from firebase_admin import credentials, auth, firestore, _auth_utils
except ModuleNotFoundError: a=1
from w3bsite.classes import defaults as _defaults_
from w3bsite.classes import email

# the users class.
class Users(_defaults_.Defaults):
	def __init__(self,
		# noreply email settings.
		email_address=None, 
		email_password=None, 
		smtp_host="smtp.gmail.com", 
		smtp_port=587, 
		# objects.
		database=None,
		firestore=None, 
		stripe=None, 
		django=None,
		logging=None, 
		# defaults.
		defaults=None,
	):	

		# defaults.
		_defaults_.Defaults.__init__(self, traceback="w3bsite.Website.users",)
		self.assign(defaults.dict())

		# variables.
		self.email_address = email_address
		self.email_password = email_password
		self.smtp_host = smtp_host 
		self.smtp_port = smtp_port 
		self.visible_email = None

		# objects.
		self.database = database
		self.firestore = firestore
		self.logging = logging
		self.stripe = stripe
		self.django = django

		# settings for saving user data.
		self.users_collection = "users/"

		# default user data.
		self.default_user_data = {
			# default data.
			"api_key":None,
			# the requests.
			"requests": {},
			# the permissions.
			"permissions": {
				"email":True,
			},
			# timestamps.
			"timestamps": {
				"signed_up":None,
			},
		}

		#
	def get(self, 
		# define one of the following user id parameters.
		username=None,
		email=None,
	):
		return self.django.users.get(username=username, email=email)
	def create(self,
		# required:
		username=None,
		email=None,
		password=None,
		verify_password=None,
		# optionals:
		name=None,
		superuser=False,
	):

		# check parameters.
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="create"),
			parameters={
				"username":username,
				"email":email,
				"password":password,
				"verify_password":verify_password,
			})

		# check password.
		password = str(password)
		verify_password = str(verify_password)
		if len(password) < 8:
			return r3sponse.error("The password must contain at least 8 characters.")
		elif password.lower() == password:
			return r3sponse.error("The password must regular and capital letters.")
		elif password != verify_password:
			return r3sponse.error("Passwords do not match.")

		# create django user.
		response = self.django.users.create(
			email=email,
			username=username,
			password=password,
			name=name,
			superuser=superuser,)
		if not response.success: return response
		user = response.user

		# save new user data.
		data = dict(self.default_user_data)
		data["timestamps"]["signed_up"] = Date().date
		data["api_key"] = String("").generate(length=68, capitalize=True, digits=True)
		_response_ = self.aes.encrypt(password)
		if not _response_.success: return self.response(_response_)
		try:data["account"]
		except: data["account"] = {}
		data["account"]["username"] = username
		data["account"]["email"] = email
		data["account"]["name"] = name
		data["account"]["password"] = _response_["encrypted"].decode()
		response = self.save_data(email, data)
		if not response.success: return response

		# insert api key cache.
		if "api_keys" in list(self.cache.__cache__.keys()):
			self.cache.__cache__["api_keys"]["value"][data["api_key"]] = {
				"email":email,
			}

		# handle success.
		return r3sponse.success(f"Successfully created user [{email}].", {
			"user":user,
			"email":user.email,
			"data":data,
		})

		#
	def update(self,
		# required:
		email=None,
		# optionals:
		name=None,
		password=None,
		verify_password=None,
		superuser=None,
		#phone_number=None,
		#photo_url=None,
		#email_verified=None,
	):

		# load.
		response = self.get(email=email)
		if response.error != None: return response
		user = response["user"]

		# check password.
		if password != None:
			if verify_password == None: return r3sponse.error("Specify parameter: verify_password.")
			password = str(password)
			verify_password = str(verify_password)
			if len(password) < 8:
				return r3sponse.error("The password must contain at least 8 characters.")
			elif password.lower() == password:
				return r3sponse.error("The password must regular and capital letters.")
			elif password != verify_password:
				return r3sponse.error("Passwords do not match.")
		
		# create django user.
		response = self.django.users.update(
			email=email,
			username=username,
			password=password,
			name=name,
			superuser=superuser,)
		if not response.success: return response
		user = response.user

		# update firestore.
		response = self.load_data(email=email)
		if not response.success: return response
		data, edits = response.data, 0
		if email != None:
			edits += 1
			data["account"]["email"] = email
		if name != None:
			edits += 1
			data["account"]["name"] = name
		if edits > 0:
			response = self.save_data(email=email, data=data)
			if not response.success: return response

		# save pass.
		if password != None:
			response = self.save_password(email=user.email, password=password)
			if not response.success: return response

		# handle success.
		return r3sponse.success(f"Successfully updated user [{email}].")
		
		#
	def delete(self, 
		# the user's email.
		email=None,
	):
		
		# check parameters.
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="delete"),
			parameters={
				"email":email,
			})
		if not response.success: return response

		# delete django user.
		response = self.django.users.delete(email=email,)
		if not response.success: return response

		# delete firestore data.
		response = self.database.delete(f"{self.users_collection}/{email}/")
		if not response.success: return response

		# handle success.
		return r3sponse.success(f"Successfully deleted user [{email}].")

		#
	def exists(self, 
		# one of the user id options is required.
		username=None,
		email=None,
		# the filter (django).
		filter="django",
	):
		if filter == "django":
			return DjangoUser.objects.filter(username=username, email=email).exists()
		else: 
			raise ValueError(f"Invalid filter [{filter}], valid options: [django].")
	def authenticate(self,
		# the users username.
		username=None,
		# the users password.
		password=None,
		# the 2fa code.
		_2fa_code=None,
		# the 2fa enabled boolean.
		# leave None to use the self._2fa settings and specify to overwrite.
		_2fa=None,
		# the request object.
		request=None,
	):

		# check parameters.
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="authenticate"),
			parameters={
				"username":username,
				"password":password,
			})
		if not response.success: return response

		# check first.
		response = self.django.users.authenticate(username=username, password=password, login=not _2fa, request=request)
		if not response.success: return response
		
		# check 2fa.		
		if _2fa == None: _2fa = self._2fa
		if _2fa == True:
			if _2fa_code == None: 

				# send code.
				l_response = self.send_verification_code(username, ip=ip, mode="authentication")
				if not l_response.success: return l_response

				# handler for sign in.
				return r3sponse.error(f"Authentication verification code required.")

			# verify code.
			response = self.verify_verification_code(username, code=_2fa_code, mode="authentication")
			if not response.success: return response

		# no 2fa.
		else:
			return response

		# success.
		return self.django.users.authenticate(username=username, password=password, request=request)

		#
	def authenticated(self, request):

		# by request user filled (django).
		if request.user.is_authenticated:
			return r3sponse.success(f"User {request.user.username} is authenticated.", {
				"api_key":None,
				"email":request.user.email,
				"username":request.user.username,
			})

		# by api key.
		optional_parameters, response = r3sponse.get_request_parameters(request, [
			"api_key",
			#"id_token",
			], optional=True)
		if optional_parameters["api_key"] != None:
			return self.verify_api_key(api_key=optional_parameters["api_key"])

		# error.
		return r3sponse.error("User is not authenticated, please specify your api key or login.")

		#
	def root_permission(self, request):

		# by request user filled (django).
		if request.user != None and request.user.is_authenticated and request.user.is_superuser:
			return r3sponse.success(f"Root permission granted ({request.user.username}).", {
				"api_key":None,
				"email":request.user.email,
				"username":request.user.username,
			})

		# by api key.
		# never root permission.

		# error.
		return r3sponse.error("Root permission denied.")

		#
	def load_data(self,
		# the user's email.
		email=None,
	):	
		# new.
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="load_data"),
			parameters={
				"email":email,
			})
		if not response.success: return response
		response = self.database.load(f"{self.users_collection}/{email}/", format="json")
		if response.error != None: return response
		return r3sponse.success(f"Successfully loaded the data of user [{email}].", {
			"data":response["document"],
		})
	def save_data(self,
		# the user's email.
		email=None,
		# the user's data.
		data={},
	):
		# new.
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="save_data"),
			parameters={
				"email":email,
			})
		if not response.success: return response
		response = self.database.save(f"{self.users_collection}/{email}/", data, format="json")
		if response.error != None: return response
		return r3sponse.success(f"Successfully saved the data of user [{email}].")
	def send_email(self, 
		# define email to retrieve user.
		email=None, 
		# the email title.
		title="Account Activation",
		# the html (str).
		html="",
	):

		# check email.
		try:
			if self.email == None:
				raise AttributeError("")
		except AttributeError:
			response = self.__initialize_email__()
			if response.error != None: return response
				
		# get user.
		response = self.get(email=email)  
		if response.error != None: return response
		user = response["user"]
		
		# send email.
		if syst3m.defaults.options.log_level >= 1:
			r3sponse.log(f"Sending email to user [{user.username}], email [{user.email}].")
		response = self.email.send(
			subject=title,
			recipients=[user.email], 
			html=html)
		if response.error != None:
			response = self.__initialize_email__()
			if response.error != None: return response
			response = self.email.send(
				subject=title,
				recipients=[user.email], 
				html=html)
			if response.error != None: return response

		# success.
		return r3sponse.success(f"Successfully send an email to user [{user.email}].")

		#
	def send_code(self, 
		# define email to retrieve user.
		email=None, 
		# the clients ip.
		ip="unknown",
		# the mode id.
		mode="verification",
		# the mode title.
		title="Account Activation",
		# the html (str).
		html="",
		# optionally specify the code (leave None to generate).
		code=None,
	):

		# check email.
		try:
			if self.email == None:
				raise AttributeError("")
		except AttributeError:
			response = self.__initialize_email__()
			if response.error != None: return response
				
		# save code.
		if code == None:
			code = Integer(0).generate(length=6)
		response = self.get(email=email)
		if response.error != None: return response
		user = response["user"]
		try: self.verification_codes[mode]
		except: self.verification_codes[mode] = {}
		self.verification_codes[mode][user.email] = {
			"code":code,
			"stamp":Date().timestamp,
			"attempts":3,
		}
		
		# send email.
		if syst3m.defaults.options.log_level >= 1:
			r3sponse.log(f"Sending [{mode}] code to user [{user.username}], email [{user.email}].")
		response = self.email.send(
			subject=f'{title} - Verification Code',
			recipients=[user.email], 
			html=html)
		if response.error != None:
			response = self.__initialize_email__()
			if response.error != None: return response
			response = self.email.send(
				subject=f'{title} - Verification Code',
				recipients=[user.email], 
				html=html)
			if response.error != None: return response

		# success.
		return r3sponse.success(f"Successfully send a verification code to user [{user.email}].")

		#
	def verify_code(self, 
		# define email to retrieve user.
		email=None, 
		# the user entered code.
		code=000000, 
		# the message mode.
		mode="verification",
	):


		# get user.
		response = self.get(email=email)
		if response.error != None: return response
		user = response["user"]

		# get success.
		fail = False
		try: self.verification_codes[mode]
		except: self.verification_codes[mode] = {}
		try: success = self.verification_codes[mode][user.email]["attempts"] > 0 and str(self.verification_codes[mode][user.email]["code"]) == str(code)
		except KeyError: fail = True

		# handle.
		if fail: 
			return r3sponse.error("Incorrect verification code.")
		elif not success: 
			self.verification_codes[mode][user.email]["attempts"] -= 1
			return r3sponse.error("Incorrect verification code.")
		else:
			del self.verification_codes[mode][user.email]
			return r3sponse.success("Successfull verification.")

		#
	def verify_api_key(self, api_key=None, request=None):
		
		# by request.
		if request != None:
			parameters, response = r3sponse.get_request_parameters(request, ["api_key"])
			if not response.success: return response
			return self.verify_api_key(api_key=parameters["api_key"])
		
		# by api key.
		else:
			
			# collect api keys from cache.
			response = self.__collect_api_keys_cache__()
			if not response.success: return response
			api_keys = response["api_keys"]

			# check api key.
			for _api_key_, info in api_keys.items():
				if _api_key_ == api_key: 
					return r3sponse.success(f"Successfully verified api key [{api_key}].", {
						"email":info["email"],
						"api_key":api_key,
					})

			# error.
			return r3sponse.error(f"Unable to verify api key [{api_key}].")

			#
	def verify_subscription(self,
		# select one of the following user id options:
		email=None,
		api_key=None,
		# the subscription product.
		product=None,
		# the subscription plans that will return a success verification (["*"] for all plans within the product).
		plans=[],
	):

		# set user id.
		if email != None:
			id = email
		elif api_key != None:
			response = self.verify_api_key(api_key=api_key)
			if not response.success: return response
			id = response["email"]
			email = response["email"]

		# collect stripe subscriptions cache.
		response = self.__collect_subscriptions_cache__()
		if not response.success: return response
		subscriptions = response["subscriptions"]

		# create user subscriptions.
		user_subscriptions = {}
		for _product_, _plans_ in self.stripe._subscriptions_.items():
			response = self.stripe.get_product_id(product=_product_)
			if not response.success: return response
			product_id = response["id"]
			user_subscriptions[_product_] = {}
			for plan_name, plan_info in _plans_.items():
				response = self.stripe.get_plan_id(product=_product_, plan=plan_name)
				if not response.success: return response
				plan_id = response["id"]
				active = False
				try: _subscriptions_ = subscriptions[email]
				except: _subscriptions_ = {}
				for _plan_id_, info in _subscriptions_.items():
					if _plan_id_ == plan_id and info["active"]:
						active = True
						break
				user_subscriptions[_product_][plan_name] = active

		# check user subscriptions.
		required_plans = list(plans)
		verified = False
		for allowed_plan in plans:
			if allowed_plan == "*":
				required_plans = list(user_subscriptions[product].keys())
				for _, allowed in user_subscriptions[product].items():
					if allowed:
						verified = True
						break
				break
			else:
				if user_subscriptions[product][allowed_plan] == True:
					verified = True
					break

		# handlers.
		if verified:
			return r3sponse.success(f"Successfully verified subscription {product} for user {id}.", {
				"email":email,
				"api_key":api_key,
				"product":product,
				"plans":required_plans,
			})
		else:
			_required_plans_ = ""
			for i in required_plans: 
				if _required_plans_ == "": _required_plans_ = i
				else: _required_plans_ += ", "+i
			return r3sponse.error(f"Permission denied ({email}). This request requires at least one of the following plans for {product} [{_required_plans_}].")

		#
	def create_subscription(self,
		# select one of the following user id options:
		email=None,
		api_key=None,
		# the subscription product.
		product=None,
		# the subscription plan.
		plan=None,
		# the card holders name.
		card_name=None,
		# the card number.
		card_number=None,
		# the card expiration month.
		card_expiration_month=None,
		# the card expiration year.
		card_expiration_year=None,
		# the card cvc.
		card_cvc=None,
	):

		# checkparameters.
		if email == None and api_key == None: return r3sponse.error("Define one of the following parameters: [email, api_key].")
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="create_subscription"),
			parameters={
				"product":product,
				"plan":plan,
			})
		if not response.success: return response
		blinded = False
		if card_cvc == "***" and "************" in card_number:
			blinded = True
		if not blinded:
			response = r3sponse.check_parameters(
				traceback=self.__traceback__(function="create_subscription"),
				parameters={
					"card_name":card_name,
					"card_number":card_number,
					"card_expiration_month":card_expiration_month,
					"card_expiration_year":card_expiration_year,
					"card_cvc":card_cvc,
				})
			if not response.success: return response
		

		# set user id.
		if email != None:
			id = email
		elif api_key != None:
			response = self.verify_api_key(api_key=api_key)
			if not response.success: return response
			id = response.email

		# get plan id.
		response = self.stripe.get_product_id(product=product)
		if not response.success: return response
		product_id = response.id
		response = self.stripe.get_plan_id(product=product, plan=plan)
		if not response.success: return response
		plan_id = response.id

		# check / create customer.
		response = self.stripe.customers.check(email=id)
		if not response.success: return response
		elif not response.exists:
			response = self.stripe.customers.create(email=id)
			if not response.success: return response
			customer_id = response.id
		else:
			customer_id = response.id

		# map user's stripe subscriptions.
		plan_ids, sub_ids, active_product_subs = [], {}, {}
		response = self.stripe.subscriptions.get(email=email)
		if not response.success and "No subscriptions found for user" not in response.error: return response
		elif not response.success and "No subscriptions found for user" in response.error:
			plan_ids, sub_ids, active_product_subs = [], {}, {}
		else:
			plan_ids, sub_ids, active_product_subs = [], {}, {}
			for _plan_id_,info in response.subscriptions.items():
				response = self.stripe.get_product_id_by_plan_id(_plan_id_)
				if not response.success: return response
				_product_id_ = response.id
				if _product_id_ == product_id:
					active_product_subs[info["plan_id"]] = info["subscription_id"]	
				plan_ids.append(info["plan_id"])
				sub_ids[info["plan_id"]] = info["subscription_id"]

		# check already subscribed.
		if plan_id in plan_ids:
			return r3sponse.error(f"User {email} is already subscribed to plan {plan} from product {product}.")

		# create card when card is not blinded (aka retrieved from user).
		if not blinded:
			response = self.stripe.customers.create_card(
				id=customer_id,
				name=card_name,
				number=card_number,
				month=card_expiration_month,
				year=card_expiration_year,
				cvc=card_cvc,)
			if not response.success: return response

		# cancel all plans within the product.	
		for _plan_id_, _subscription_id_ in active_product_subs.items():
			response = self.stripe.subscriptions.cancel(subscription_id=_subscription_id_)
			if not response.success: return response		

		# create subscription.
		response = self.stripe.subscriptions.create(customer_id=customer_id, plans=[plan_id])
		if not response.success: return response
		
		# handlers.
		return r3sponse.success(f"Successfully created subscription {product} {plan} for user {id}.")

		#
	def get_api_key(self, email=None, username=None):

		# collect api keys from cache.
		response = self.__collect_api_keys_cache__()
		if not response.success: return response
		api_keys = response["api_keys"]

		# get api key.
		api_key, id = None, None
		if email != None: id = email
		elif username != None: 
			response = self.get(username=username)
			if not response.success: return response
			id = response["user"].email
		else: return r3sponse.error("Define parameter email or username.")
		for _api_key_, info in api_keys.items():
			if info["email"] == id: 
				api_key = _api_key_
				break
		if api_key != None:
			return r3sponse.success(f"Successfully retrieved the api key [{id}].", {
				"api_key":api_key,
			})
		else:
			return r3sponse.error(f"Unable to find api key [{id}].")

		#
	def set_permission(self, email, permission_id, permission=True):
		
		# load user data.
		response = self.load_data(email=email) 
		if not response.success: return response
		data = response["data"]

		# edit data.
		data["permissions"][permission_id] = permission

		# save data.
		response = self.save_data(email=email, data=data) 
		if response.error != None: return response

		# handlers.
		return r3sponse.success(f"Successfully set the {permission_id} permission of user [{email}] to [{permission}].")
		#
	def load_password(self, email=None):
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="load_password"),
			parameters={
				"email":email,
			})
		if not response.success: return response
		response = self.load_data(email)
		if not response.success: return response
		data = response.data
		_response_ = self.aes.decrypt(data["account"]["password"])
		if not _response_.success: return self.response(_response_)
		password = _response_["decrypted"].decode()
		return r3sponse.success(f"Successfully retrieved the password of user {email}.",{
			"password":password,
			"data":data,
		})
	def save_password(self, email=None, password=None):
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="save_password"),
			parameters={
				"email":email,
				"password":password,
			})
		if not response.success: return response
		response = self.load_data(email)
		if not response.success: return response
		data = response.data
		_response_ = self.aes.encrypt(password)
		if not _response_.success: return self.response(_response_)
		try:data["account"]
		except: data["account"] = {}
		data["account"]["password"] = _response_["encrypted"].decode()
		response = self.save_data(email, data)
		if not response.success: return response
		return r3sponse.success(f"Successfully saved the password of user {email}.")
	def iterate(self, emails=False, firestore=False, combined=False):
		if combined:
			firestore = self.iterate(emails=emails, firestore=True)
			django = self.iterate(emails=emails, firestore=False)
			return firestore + django
		else:
			if firestore:
				response = self.firestore.load_collection("users/")
				if not response.success: raise ValueError(response.error)
				users = []
				_emails_ = response["collection"]
				if not emails:
					for email in _emails_:
						users.append(auth.get_user_by_email(email))
					return users
				else:
					return _emails_
			else:
				users, _emails_ = auth.list_users().iterate_all(), []
				if emails:
					for user in users:
						_emails_.append(user.email)
					return _emails_
				else:
					return users
	def synchronize(self, 
		# leave emails=None default to synchronize all users.
		# optionally pass emails=[newuser@email.com] to synchronize new users.
		emails=None, 
	):

		# get all emails.
		if not isinstance(emails, list):
			emails = self.iterate(emails=True, firestore=True)

		# iterate.
		for email in emails: 

			# check django user existance (in case of sql rebuild).
			response = self.django.users.exists(email=email)
			if not response.success: return response
			elif not response.exists:
				response = self.load_password(email=email)
				if not response.success: return response
				data, password = response.data, response.password
				response = self.django.users.create(
					email=email, 
					password=password, 
					username=data["account"]["username"], 
					name=data["account"]["name"])
				if not response.success: return response

			# get user.
			response = self.get(email=email)
			if not response.success: return response
			user = response["user"]

			# load data.
			response = self.load_data(email=email)
			if not response.success: return response
			data, edits = response["data"], 0

			# check user data.
			d = Dictionary(dictionary=data, path=False)
			old = dict(data)
			data = d.check(default=self.default_user_data)
			if old != data:
				edits += 1

			# check user data.
			try:
				if data["api_key"] in [None, ""]: raise KeyError("")
			except KeyError: 
				data["api_key"] = String("").generate(length=68, capitalize=True, digits=True)
				edits += 1

			# edits.
			if edits > 0:
				response = self.save_data(email=email, data=data)
				if not response.success: return response

		# success.
		return r3sponse.success(f"Successfully synchronized {len(emails)} user(s).")

		#
	# system functions.
	def __initialize_email__(self):
		# the email object.
		if self.email_address == None or self.email_password == None:
			return r3sponse.error("Define the firebase.users.email_address & firebase.users.email_password variables to send emails.")
		self.email = email.Email(
			email=self.email_address,
			password=self.email_password,
			smtp_host=self.smtp_host,
			smtp_port=self.smtp_port,
			visible_email=self.visible_email,)
		response = self.email.login()
		if response["success"]:
			return r3sponse.success("Successfully initialized the mail object.")
		else: 
			self.email = None
			return response
	def __collect_subscriptions_cache__(self, refresh=False):

		# get cache.
		subscriptions, cache_subscriptions = {}, False
		response = self.cache.get(key="subscriptions")
		if not response.success: 
			cache_subscriptions = True
		else:
			subscriptions = response["value"]

		# collect cache.
		if refresh or cache_subscriptions:
			response = self.stripe.subscriptions.get(active_only=True, by_customer_id=False)
			if not response.success: return response
			subscriptions = response["subscriptions"]

		# set cache.
		if refresh or cache_subscriptions:
			response = self.cache.set(key="subscriptions", value=subscriptions, reset=60)
			if not response.success: return response

		# handler.
		return r3sponse.success("Successfully retrieved the stripe subscriptions cache.", {
			"subscriptions":subscriptions,
		})

		#
	def __collect_api_keys_cache__(self, refresh=False):
		
		# get cache.
		api_keys, cache_api_keys = {}, False
		response = self.cache.get(key="api_keys")
		if not response.success: 
			cache_api_keys = True
		else:
			api_keys = response["value"]

		# collect cache.
		if cache_api_keys or refresh:
			api_keys = {}
			for email in self.iterate(emails=True, combined=True):
				response = self.load_data(email=email)
				if not response.success: return response
				_api_key_ = response["data"]["api_key"]
				api_keys[_api_key_] = {
					"email":email,
				}

		# set cache.
		if cache_api_keys or refresh:
			response = self.cache.set(key="api_keys", value=api_keys, reset=60*24)
			if not response.success: return response

		# handler.
		return r3sponse.success("Successfully retrieved the api keys cache.", {
			"api_keys":api_keys,
		})

		#

	


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils

# check installation.
"""
if os.environ.get("DJANGO_SECRET_KEY") == None and not Files.exists("/usr/local/bin/heroku"):
	if OS in ["macos"]:
		os.system("brew tap heroku/brew && brew install heroku")
	else:
		os.system("curl https://cli-assets.heroku.com/install.sh | sh")
	if not Files.exists("/usr/local/bin/heroku"):
		raise ImportError("Failed to install the heroku cli.")
"""

# the heroku object class.
class Heroku(object):
	def __init__(self, 
		# the root path.
		root=None,
		# the domain.
		doman=None,
		# the website name.
		name=None,
		# passed objects.
		namecheap=None,
		logging=None,
	):

		# arguments.
		self.root = root
		self.domain = domain
		self.sub_domain = sub_domain
		self.name = name
		self.namecheap = namecheap
		self.logging = logging
		response = self.check_logged_in()
		if response.error != None: raise RuntimeError(response.error)

		#
	# x.
	def check(self):
		os.system("cd "+self.root+"/" + " && " + "heroku ps:scale web=1" + " && " + "./manage.py makemigrations" + " && " + "./manage.py migrate")
	def tail(self):
		os.system("cd "+self.root+"/" + " && " + "heroku logs --tail")
	def add_environment_variables(self, variables={}, silent=True):
		c = "cd "+self.root+"/"
		c += " && heroku config:set"
		for key, value in variables.items():
			c += " {}='{}'".format(key, str(value).replace("\n","\\n"))
		output = utils.__execute_script__(c)
		success = True
		for key, _ in variables.items():
			if key+":" not in output: success = False ; break
		if success:
			return r3sponse.success("Success")
		else:
			return r3sponse.error(f"Error, output: {output}")
	def remove_environment_variables(self, variables={}):
		c = "cd "+self.root+"/"
		for key, value in variables.items():
			c += "&&" + f"heroku config:unset {key}"
		output = utils.__execute_script__(c)
		print(output)
	def get_environment_variables(self, variables={}):
		c = "cd "+self.root+"/"
		for key, value in variables.items():
			c += "&&" + f"heroku config:get {key}"
		output = utils.__execute_script__(c)
		print(output)
	# x.
	def push(self, log_level=0):
		output = utils.__execute_script__(f"""
		cd {self.root}/ && 
		git add . && 
		git commit -am 'automatic updates' && 
		git push heroku master
		""", silent=log_level!=0)
		if "Verifying deploy... done." in output:
			return r3sponse.success(f"Successfully pushed website [{self.root}] to heroku.")
		elif "Everything up-to-date" in output or "nothing to commit, working tree clean" in output:
			return r3sponse.error(f"Website [{self.root}] is already up to date.")
		else:
			return r3sponse.error(f"Failed to push website [{self.root}] to heroku.")
	def get_deploy_app(self):

		# variables.
		domains = {}

		# get.
		output = utils.__execute_script__(f"""
			cd {self.root}/
			heroku domains
		""", return_format="array")

		# iterate.
		app, mode, domain = None, None, None
		new = False
		for line in output:
			if line not in ["", " "]:

				# set app and mode.
				if line[:len("=== ")] == "=== ":
					l_app = line.split("=== ")[1].split(" ")[0]
					mode = line.split(f"=== {l_app} ")[1].split("\n")[0]
					try: domains[l_app]
					except KeyError: domains[l_app] = {}
					if app == None:
						app = str(l_app)
					elif mode == "Heroku Domain":
						return r3sponse.error(f"Detected multiple heroku apps for project [{self.root}].")				

				
		# handlers.
		return r3sponse.success(f"Successfully retrieved the heroku app name of project [{self.root}].", {"app":app})

		#
	def get_deploy_domain(self,
		# the heroku app name (optional to increase speed).
		app=None,
	):

		# get info.
		if app == None:
			response = self.get_deploy_app()
			if response.error != None: return response
			app = response["app"]
		response = self.get_domains()
		if response.error != None: return response

		# handlers.
		try:
			dict = response["domains"][app]
		except KeyError:
			return r3sponse.error(f"Unable to find heroku app [{app}] from project [{self.root}.")
		if len(dict) == 0:
			return r3sponse.error(f"Unable to find any heroku domains for heroku app [{app}] from project [{self.root}].")
		elif len(dict) > 1:
			return r3sponse.error(f"Found multiple heroku domains for heroku app [{app}] from project [{self.root}].")
		else:
			return r3sponse.success(f"Successfully found the deploy domain for project [{self.root}].", {
				"domain":list(dict.keys())[0]
			})

		#
	def get_domains(self):

		# variables.
		domains = {}

		# get.
		output = utils.__execute_script__(f"""
			cd {self.root}/
			heroku domains
		""", return_format="array")

		# iterate.
		app, mode, domain = None, None, None
		for line in output:
			if line not in ["", " "]:

				# set app and mode.
				if line[:len("=== ")] == "=== ":
					app = line.split("=== ")[1].split(" ")[0]
					mode = line.split(f"=== {app} ")[1].split("\n")[0]
					try: domains[app]
					except KeyError: domains[app] = {}
				
				# by mode.
				elif mode == "Heroku Domain":
					domain = line.split('\n')[0]
					try: domains[app][domain]
					except KeyError: domains[app][domain] = {}
				elif mode == "Custom Domains":
					if "Domain Name" not in line and "DNS" not in line and "Record Type" not in line and "Target" not in line:
						custom_domain = line.split(" ")[0]
						for i in range(101):
							if "  " not in line: break
							line = line.replace("  "," ")
						if line[len(line)-1] == " ": line = line[:-1]
						target = line.split(" ")[len(line.split(" "))-1]
						if target[len(target)-1] == ".": target = target[:-1]
						domains[app][domain][custom_domain] = target

		# handlers.
		return r3sponse.success(f"Successfully retrieved the heroku domains of project [{self.root}].", {"domains":domains})

		#
	def check_domain(self, domain=None):
		# set default domain.
		if domain == None: domain = self.domain
		response = self.get_domains()
		if response.error != None: return response
		match = False
		for app, info in response["domains"].items():
			for heroku_domain, info_i in info.items():
				for _domain_, target in info_i.items():
					if _domain_ == domain:
						match = True
						break
		return r3sponse.success(f"Successfully checked the existance of domain [{domain}].", {"exists":match})
	def add_domain(self, domain=None):

		# set default domain.
		if domain == None: domain = self.domain

		# check duplicates.
		response = self.check_domain(domain)
		if response.error != None: return response
		elif response["exists"]:
			return r3sponse.error(f"Domain [{domain}] is already added to heroku project [{self.root}].")

		# add.
		output = utils.__execute_script__(f"""
			cd {self.root}
			heroku domains:add {domain}
		""", return_format="array")

		# check existsnce.
		response = self.check_domain(domain)
		if response.error != None: return response
		elif response["exists"]:
			return r3sponse.success(f"Successfully added domain [{domain}] to heroku project [{self.root}].")
		else:
			return r3sponse.error(f"Failed to add domain [{domain}] to heroku project [{self.root}].")

		#
	def check_logged_in(self):
		output = utils.__execute_script__(f"""
			cd {self.root}
			heroku auth:whoami
		""")
		if "Error: not logged in" in output:
			return r3sponse.error("The heroku cli is not signed in. Open a terminal and execute the following command to login: [ $ heroku login].")
		else:
			return r3sponse.success("The heroku cli is Successfully signed in.")

		#	
	def install_tls(self,
		# the heroku app name (optional to increase speed).
		app=None,
	):
		#source: https://www.namecheap.com/support/knowledgebase/article.aspx/9756/33/installing-an-ssl-certificate-on-heroku-paid-ssl-endpoint

		# get app name.
		if app == None:
			response = self.get_deploy_app()
			if response.error != None: return response
			app = response["app"]

		# add.
		output = utils.__execute_script__(f"""
			cd {self.root}
			heroku certs:add .secrets/tls/server.crt .secrets/tls/server.key
		""")

		# update.
		if "try certs:update instead" in output:
			output = utils.__execute_script__(f"""
				cd {self.root}
				heroku certs:update --confirm {app} .secrets/tls/server.crt .secrets/tls/server.key
			""")

		# handlers.
		if "Your certificate has been added successfully." in output or ("Updating SSL certificate " in output and "... done" in output):
			return r3sponse.success(f"Successfully installed the tls certificate of domain [{domain}].")
		else:
			return r3sponse.error(f"Failed to install the tls certificate of domain [{domain}].")
	def check_dns(self, log_level=0):

		# check heroku (sub) domains.
		domains = [self.domain]
		for domain in domains:
			response = self.check_domain(domain)
			if response.error != None: 
				r3sponse.log(response=response, log_level=log_level)
				return response
			elif not response["exists"]:
				loader = syst3m.console.Loader(f"Adding domain [{domain}].")
				response = self.add_domain(domain)
				loader.stop(response=response)
				if response.error != None: 
					r3sponse.log(response=response, log_level=log_level)
					return response

		# get heroku app.
		response = self.get_domains()
		if response.error != None: 
			r3sponse.log(response=response, log_level=log_level)
			return response
		heroku_domains = response["domains"]

		# get heroku app.
		response = self.get_deploy_app()
		if response.error != None: 
			r3sponse.log(response=response, log_level=log_level)
			return response
		app = response["app"]
		print(f"Heroku deploy app: {app}")

		# get heroku deploy domain.
		response = self.get_deploy_domain(app=app)
		if response.error != None: 
			r3sponse.log(response=response, log_level=log_level)
			return response
		deploy_domain = response["domain"]
		print(f"Heroku deploy domain: {deploy_domain}")

		# check dns settings.
		for domain in domains:

			# check namecheap domain.
			response = self.namecheap.check_domain(domain)
			if response.error != None: 
				r3sponse.log(response=response, log_level=log_level)
				return response
			elif not response["exists"]:
				return r3sponse.error(f"Specified domain [{domain}] is not owned by namecheap user [{self.namecheap.username}].", log_level=log_level)

			# get dns target.
			try:
				target = heroku_domains[app][deploy_domain][domain]
			except KeyError:
				return r3sponse.error(f"Specified domain [{domain}] is not added to the domains of heroku app [{app}] from project [{self.name}].", log_level=log_level)
			print(f"Heroku DNS target: {target}")

			# add dns records.
			response = self.namecheap.add_dns(
				# the domain (optional).
				domain=domain,
				# the dns record type,
				type="CNAME",
				# the dns record host,
				host="www",
				# the dns record value/address,
				value=target,
				# the get_dns.dns dictionary (optionally to increase speed).
				records=None,)
			if response.error != None and "] already exists." not in response.error: 
				r3sponse.log(response=response, log_level=log_level)
				return response
			elif response.error == None: 
				r3sponse.log(response=response, log_level=log_level)
			response = self.namecheap.add_dns(
				# the domain (optional).
				domain=domain,
				# the dns record type,
				type="ALIAS",
				# the dns record host,
				host="@",
				# the dns record value/address,
				value=target,
				# the get_dns.dns dictionary (optionally to increase speed).
				records=None,)
			if response.error != None and "] already exists." not in response.error: 
				r3sponse.log(response=response, log_level=log_level)
				return response
			elif response.error == None: 
				r3sponse.log(response=response, log_level=log_level)
		
		# handlers.
		return r3sponse.success(f"Successfully checked the heroku dns settings for domain {self.domain}.", log_level=log_level)

		#
	def deploy(self, log_level=0):

		# set heroku env.
		loader = syst3m.console.Loader("Configuring heroku environment ...")
		response = self.add_environment_variables(variables=utils.__load_json__(f"{self.root}/.secrets/env.json"))
		if not response.success: 
			loader.stop(success=False)
			return response
		loader.stop()

		# push to heroku.
		loader = syst3m.console.Loader("Pushing to heroku ...")
		response = self.push(log_level=log_level)
		if response.error != None and "] is already up to date." not in response.error: 
			loader.stop(success=False)
			return response
		elif response.error == None:
			loader.stop()
			r3sponse.log(response=response, log_level=log_level)

		# handlers.
		return r3sponse.success(f"Successfully deployed website {self.domain} on heroku.", log_level=log_level)

		#

#


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *

# the website defaults object class.
class Defaults(syst3m.objects.Object):
	def __init__(self,
		# info.
		root=None,
		library=None,
		database_path=None,
		name=None,
		domain=None,
		https_domain=None,
		author=None,
		email=None,
		country_code=None,
		province=None,
		city=None,
		organization=None,
		organization_unit=None,
		developers=None,
		remote=None,
		live=None,
		interactive=None,
		_2fa=None,
		maintenance=False,
		# objects.
		template_data=None,
		aes=None,
		cache=None,
		logging=None,
		# defaults.
		traceback="w3bsite.Website.defaults",
	):

		# object defaults.
		syst3m.objects.Object.__init__(self, traceback=traceback)

		# defaults.
		self.root = root
		self.library = library
		self.database_path = database_path
		self.name = name
		self.author = author
		self.email = email
		self.organization = organization
		self.country_code = country_code
		self.domain = domain
		self.https_domain = https_domain
		self.developers = developers
		self.remote = remote
		self.live = live
		self.interactive = interactive
		self.province = province
		self.city = city
		self.organization_unit = organization_unit
		self._2fa = _2fa
		self._maintenance_ = maintenance

		# objects.
		self.template_data = template_data
		self.aes = aes
		self.cache = cache
		self.logging = logging

		#

	


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils
from w3bsite.classes import defaults as _defaults_

# pip imports.
from django.contrib.auth import authenticate, login
import firebase_admin
from firebase_admin import credentials, auth, firestore, _auth_utils

class FirebaseCLI(object):
	def __init__(self):
		a=1
	"""
		if "--install" in sys.argv:
			if not self.installed():
				self.install():
	def install(self):
		os.system("curl -sL https://firebase.tools | bash")
	def installed(self):
		return Files.exists("/usr/local/bin/firebase") or Files.exists("/usr/bin/firebase")
	def login(self):
		output = util.__execute_script__("firebase login")
	def projects(self):
		output = util.__execute_script__("firebase")
	"""
	
# the firebase class.
class Firebase(_defaults_.Defaults):
	def __init__(self, 
		# the firebase key.
		key=None, 
		# the firebase js config.
		firebase_js={},
		# defaults.
		defaults=None,
	):
		
		# defaults.
		_defaults_.Defaults.__init__(self)
		self.assign(defaults.dict())

		# check arguments.
		#response = r3sponse.check_parameters({
		#	#"ip":ip,
		#})
		#if not response.success: raise ValueError(response.error)
		

		# initialize firestore.
		# (in classes.config)
		try:
			cred = credentials.Certificate(key) # must still be edited through env variables.
			firebase_admin.initialize_app(cred)
		except ValueError:a=1
		self.firestore = FireStore()
		self.users = Users(defaults=defaults, firestore=self.firestore)

		# firebase template data.
		def __handle__(dictionary={}):
			new_dictionary = {}
			for key, value in dictionary.items():
				if isinstance(value, str) or isinstance(value, int) or value == None:
					new_dictionary[key.upper()] = value
				elif isinstance(value, dict):
					new_dictionary[key.upper()] = __handle__(value)
				else:
					raise ValueError(f"Cannot serialize template data from firebase_js key [{key}], value: [{value}].")
			return new_dictionary
		self.template_data["FIREBASE"] = __handle__(firebase_js)
		#

# the firestore class.
class FireStore(object):
	def __init__(self):
		
		# initialize firestore.
		self.db = firestore.client()

		#
	# system functions.
	def list(self, reference):
		doc = self.__get_doc__(reference)
		try:
			doc = doc.get()
			success = True
		except: success = False
		if not success:
			return r3sponse.error(f"Failed to load document [{reference}].")
		if not isinstance(doc, list):
			return r3sponse.error(f"Reference [{reference}] leads to a document, not a collection.")
		return r3sponse.success(f"Successfully listed the content of collection [{reference}].", {"collection":doc})
	def load(self, reference):
		doc = self.__get_doc__(reference)
		try:
			doc = doc.get()
			success = True
		except: success = False
		if not success:
			return r3sponse.error(f"Failed to load document [{reference}].")
		if isinstance(doc, list):
			return r3sponse.error(f"Reference [{reference}] leads to a collection, not a document.")
		if not doc.exists:
			return r3sponse.error(f"Document [{reference}] does not exist.")
		else:
			data = doc.to_dict()
			return r3sponse.success(f"Successfully loaded document [{reference}].", {"document":data})
	def load_collection(self, reference):
		doc = self.__get_doc__(reference)
		try:
			doc = doc.get()
			success = True
		except: success = False
		if not success:
			return r3sponse.error(f"Failed to load document [{reference}].")
		if isinstance(doc, dict):
			return r3sponse.error(f"Reference [{reference}] leads to a document, not a collection.")
		data = []
		for i in doc:
			if i.exists:
				data.append(i.id)
		return r3sponse.success(f"Successfully loaded the document names of collection [{reference}].", {"collection":data, "documents":doc})
	def save(self, reference, data):
		doc = self.__get_doc__(reference)
		try:
			doc.set(data)
			success = True
		except: success = False
		if success:
			return r3sponse.success(f"Successfully saved document [{reference}].")
		else:
			return r3sponse.error(f"Failed to save document [{reference}].")
	def delete(self, reference):
		doc = self.__get_doc__(reference)
		try:
			doc.delete()
			success = True
		except: success = False
		if success:
			return r3sponse.success(f"Successfully deleted document [{reference}].")
		else:
			return r3sponse.error(f"Failed to delete document [{reference}].")
	# system functions.
	def __get_doc__(self, reference):
		reference = reference.replace("//", "/")
		if reference[len(reference)-1] == "/": reference = reference[:-1]
		doc, c = None, 0
		for r in reference.split("/"):
			if doc == None:
				doc = self.db.collection(r)
				c = 1
			else:
				if c == 1:
					doc = doc.document(r)
					c = 0
				else:
					doc = doc.collection(r)
					c = 1
		return doc

# the firebase users class.
# firebase users are no longer created, just in firestore database & django.
class Users(_defaults_.Defaults):
	def __init__(self, defaults=None, firestore=None):
		
		# defaults.
		_defaults_.Defaults.__init__(self)
		self.assign(defaults.dict())

		# objects.
		self.firestore = firestore

		# variables.
		self.users_collection = "users/"

		#
	def get(self, 
		# define one of the following parameters.
		uid=None,
		email=None,
		phone_number=None,
	):
		try:
			user, variable = None, None
			if uid not in [None, "None", ""]:
				user = auth.get_user(uid)
				variable = str(uid)
			elif email not in [None, "None", ""]:
				user = auth.get_user_by_email(email)
				variable = str(email)
			elif phone_number not in [None, "None", ""]:
				user = auth.get_user_by_phone_number(phone_number)
				variable = str(phone_number)
			else:
				return r3sponse.error("Invalid usage, define one of the following parameters: [uid, email, phone_number].")
		except _auth_utils.UserNotFoundError:
			return r3sponse.error("User does not exist.")

		# check success.
		if user == None: 
			return r3sponse.error(f"Failed to retrieve user [{variable}].")
		else:
			return r3sponse.success(f"Successfully retrieved user [{variable}].", {"user":user})


		#
	def create(self,
		# required:
		email=None,
		password=None,
		verify_password=None,
		# optionals:
		name=None,
		phone_number=None,
		photo_url=None,
		email_verified=False,
	):

		# check parameters.
		response = r3sponse.check_parameters(empty_value=None, parameters={
			"email":email,
			"password":password,
			"verify_password":verify_password,
		})

		# check password.
		password = str(password)
		verify_password = str(verify_password)
		if len(password) < 8:
			return r3sponse.error("The password must contain at least 8 characters.")
		elif password.lower() == password:
			return r3sponse.error("The password must regular and capital letters.")
		elif password != verify_password:
			return r3sponse.error("Passwords do not match.")

		# create.
		try:
			user = auth.create_user(
				email=email,
				email_verified=email_verified,
				phone_number=phone_number,
				password=password,
				display_name=name,
				photo_url=photo_url,
				disabled=False)
			success = True
		except Exception as e: 
			success = False
			error = e

		# handle error.
		if not success:
			return r3sponse.error(f"Failed to create user [{email}], error: {error}")

		# handle success.
		return r3sponse.success(f"Successfully created user [{email}].", {
			"user":user,
			"uid":user.uid,
			"email":user.email,
		})

		#
	def update(self,
		# required:
		email=None,
		# optionals:
		name=None,
		password=None,
		verify_password=None,
		phone_number=None,
		photo_url=None,
		email_verified=None,
	):

		# load.
		response = self.get(email=email)
		if response.error != None: return response
		user = response["user"]
		uid = user.uid

		# set defaults.
		if name == None: 
			#name = user.display_name # firebase
			name = user.display_name # firebase
		if email == None: email = user.email
		if phone_number == None: phone_number = user.phone_number
		if photo_url == None: photo_url = user.photo_url
		if email_verified == None: email_verified = user.email_verified
		if password != None and verify_password != None:
			# check password.
			password = str(password)
			verify_password = str(verify_password)
			if len(password) < 8:
				return r3sponse.error("The password must contain at least 8 characters.")
			elif password.lower() == password:
				return r3sponse.error("The password must regular and capital letters.")
			elif password != verify_password:
				return r3sponse.error("Passwords do not match.")

		# create
		try:
			user = auth.update_user(
				uid,
				email=email,
				phone_number=phone_number,
				email_verified=email_verified,
				password=password,
				display_name=name,
				photo_url=photo_url,
				disabled=False)
			success = True
		except Exception as e: 
			success = False
			error = e

		# handle success.
		if success:
			return r3sponse.success(f"Successfully updated user [{uid}].")
		else:
			return r3sponse.error(f"Failed to update user [{uid}], error: {error}")

		#
	def delete(self, 
		# option 1:
		# the user's uid (much faster).
		uid=None,
		# option 2:
		# the users email / username.
		email=None,
	):
		if uid != None:
			try:
				auth.delete(uid)
				success = True
			except Exception as e: 
				success = False
				error = e
			if not success:
				return r3sponse.error(f"Failed to delete user [{uid}], error: {error}")
			response = self.firestore.delete(f"{self.users_collection}/{uid}")
			if not response.success: return response
			return r3sponse.success(f"Successfully deleted user [{uid}].")
		else:
			response = self.get(email=email)
			if not response.success: return response
			return self.delete(uid=response.user.uid)
	def verify_id_token(self, id_token):
		"""
			Javascript:
				firebase.auth().currentUser.getIdToken(/* forceRefresh */ true).then(function(idToken) {
				  // Send token to your backend via HTTPS
				  // ...
				}).catch(function(error) {
				  // Handle error
				});
		"""
		try:
			decoded_token = auth.verify_id_token(id_token)
			uid = decoded_token['uid']
			if uid == None: success = False
			else: success = True
		except Exception as e: 
			success = False
			error = e
		if not success:
			return r3sponse.error(f"You are not signed in, error: {error}")
		response = self.get(uid=uid)
		if not response.success: return response
		user = response.user
		return r3sponse.success("You are signed in.", {"uid":uid, "email":user.email})	

		#
		





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *

# the stripe subscription daemon.
# used for verifying subscriptions.
class SubscriptionVerification(syst3m.objects.Thread):
	def __init__(self, 
		# the daemons sleep time.
		sleeptime=60,
		# the website.users object,.
		users=None,
		# the website.stripe object.
		stripe=None, 
	):
		
		# defaults.
		syst3m.objects.Thread.__init__(self, log_level=0)
		self.assign(defaults.dict())

		# arguments.
		self.sleeptime = sleeptime

		# objects.
		self.users = users
		self.stripe = stripe

		#
	def __run__(self):
		while self.running:
			if self.should_stop(): break
			time.sleep(self.sleeptime)
	def verify(self,
		# the users email.
		email=None,
		# the users api key.
		api_key=None,
	):
		a=1
		


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils

# the namecheap object class.
class Namecheap(syst3m.objects.Object):
	# does not use Defaults so also accessable without Website init.
	# source: https://www.namecheap.com/support/api/methods/
	# go to [Profile > Tools > Developer > API > Manage] and enable the APi and whitelist your public ip.
	def __init__(self,
		# your namecheap username.
		username=None,
		# your namecheap api key.
		api_key=None,
		# the root path.
		root=None,
		# the domain.
		domain=None,
		# the organization's email.
		email=None,
		# sandbox boolean (does not seem to work namecheap end).
		sandbox=False,
	):	

		# defaults.
		syst3m.objects.Object.__init__(self, traceback="w3bsite.Website.namecheap",)
		
		# arguments.
		self.username = username
		self.api_key = api_key
		self.domain = domain
		self.root = root
		self.domain = domain
		self.email = email

		# pre domain.
		self.pre_domain, splitted = "", self.domain.split(".")
		if len(splitted) > 2:
			c, stop = 0, len(splitted)-1
			for i in splitted:
				c += 1
				if c == stop: break
				if self.pre_domain == "": self.pre_domain = i
				else: self.pre_domain += "."+i
		self.post_domain = self.domain.replace(f"{self.pre_domain}.", "")

		# get public ip.
		self.public_ip = NETWORK_INFO["public_ip"]

		# set api base url.
		if sandbox:
			self.base_url = f"https://api.sandbox.namecheap.com/xml.response?ApiUser={self.username}&ApiKey={self.api_key}&UserName={self.username}&ClientIp={self.public_ip}"
		else:
			self.base_url = f"https://api.namecheap.com/xml.response?ApiUser={self.username}&ApiKey={self.api_key}&UserName={self.username}&ClientIp={self.public_ip}"
		
		# set tld & sld.
		self.sld, self.tld = None, None
		if self.domain != None:
			response = self.get_sld_and_tld(self.domain)
			if response.error != None: raise ValueError(response['error'])
			self.sld, self.tld = response["sld"], response["tld"]

		#
	def check_domain(self, domain=None):

		# set default domain.
		if domain == None: domain = self.domain

		# get domains.
		response = self.get_domains()
		if response.error != None: return response
		try:
			info = response["domains"][domain]
			exists = True
		except: exists = False
		return r3sponse.success(f"Successfully checked the ownership for domain [{domain}] from namecheap user [{self.username}].", {"exists":exists})

		#
	def get_domains(self):

		# api request.
		response = self.__request__("namecheap.domains.getList")
		if response["api_response"]["errors"] != None: 
			try: error = response["api_response"]["errors"]["error"]["#text"]
			except KeyError: error = response["api_response"]["errors"]
			return r3sponse.error(f"Failed to retrieve the domains of namecheap user [{self.username}], error: {error}")

		# handle response.
		domains = {}
		try: result = response["api_response"]["command_response"]["domain_get_list_result"]
		except KeyError: result = {}
		for key, domain in result.items():
			if isinstance(domain, dict):
				domains[domain["name"]] = domain
			elif isinstance(domain, list):
				for item in domain:
					item = self.__serialize_dictionary__(item)
					domains[item["name"]] = item
			else: raise ValueError(f"Unkown domain type: {domain}.")

		# handlers.
		return r3sponse.success(f"Successfully retrieved the domains of namecheap user [{self.username}].", {
			"domains":domains,
		})

		#
	def get_info(self, domain=None):

		# set default domain.
		if domain == None: domain = self.domain

		# check domain.
		response = self.check_domain(domain)
		if response.error != None: return response
		elif not response["exists"]:
			return r3sponse.error(f"Namecheap user [{self.username}] does not own domain [{domain}].")

		# api request get info.
		response = self.__request__("namecheap.domains.getInfo", {"DomainName":domain})
		if response["api_response"]["errors"] != None: 
			try: error = response["api_response"]["errors"]["error"]["#text"]
			except KeyError: error = response["api_response"]["errors"]
			return r3sponse.error(f"Failed to retrieve the info of domain [{domain}] from namecheap user [{self.username}], error: {error}")

		# handle response.
		try:
			info = response["api_response"]["command_response"]["domain_get_info_result"]
		except KeyError:
			return r3sponse.error(f"Failed to retrieve the info of domain [{domain}] from namecheap user [{self.username}].")

		# handlers.
		return r3sponse.success(f"Successfully retrieved the domain info of namecheap user [{self.username}].", {
			"info":info,
		})

		#
	def get_dns(self, domain=None):

		# set default domain.
		if domain == None: domain = self.domain

		# check domain.
		response = self.check_domain(domain)
		if response.error != None: return response
		elif not response["exists"]:
			return r3sponse.error(f"Namecheap user [{self.username}] does not own domain [{domain}].")

		# api request.
		response = self.get_sld_and_tld(domain)
		if response.error != None: raise ValueError(response['error'])
		sld, tld = response["sld"], response["tld"]
		response = self.__request__("namecheap.domains.dns.getHosts", {"SLD":sld, "TLD":tld})
		if response["api_response"]["errors"] != None: 
			try: error = response["api_response"]["errors"]["error"]["#text"]
			except KeyError: error = response["api_response"]["errors"]
			return r3sponse.error(f"Failed to retrieve the dns info for domain [{domain}] from namecheap user [{self.username}], error: {error}")

		# handle response.
		try:
			data = response["api_response"]["command_response"]["domain_dnsget_hosts_result"]
		except KeyError:
			return r3sponse.error(f"Failed to retrieve the dns info for domain [{domain}] from namecheap user [{self.username}].")

		# iterate data.
		records = {}
		if isinstance(data["host"], dict):
			serialized = self.__serialize_dictionary__(data["host"])
			if serialized['address'][len(serialized['address'])-1] == ".": serialized['address'] = serialized['address'][:-1]
			tag = self.tag_dns(
				# the dns record type,
				type=serialized['type'],
				# the dns record host,
				host=serialized['name'],
				# the dns record value/address,
				value=serialized['address'],)["tag"]
			old_id, new_id = "name", 'host' ; i = serialized[old_id] ; del serialized[old_id] ; serialized[new_id] = i
			old_id, new_id = "address", 'value' ; i = serialized[old_id] ; del serialized[old_id] ; serialized[new_id] = i
			records[tag] = serialized
		elif isinstance(data["host"], list):
			for item in data["host"]:
				serialized = self.__serialize_dictionary__(item)
				if serialized['address'][len(serialized['address'])-1] == ".": serialized['address'] = serialized['address'][:-1]
				tag = self.tag_dns(
					# the dns record type,
					type=serialized['type'],
					# the dns record host,
					host=serialized['name'],
					# the dns record value/address,
					value=serialized['address'],)["tag"]
				old_id, new_id = "name", 'host' ; i = serialized[old_id] ; del serialized[old_id] ; serialized[new_id] = i
				old_id, new_id = "address", 'value' ; i = serialized[old_id] ; del serialized[old_id] ; serialized[new_id] = i
				records[tag] = serialized

		# handlers.
		return r3sponse.success(f"Successfully retrieved the dns info of namecheap user [{self.username}].", {
			"records":records,
		})

		#
	def check_dns(self,
		# the domain (optional).
		domain=None,
		# the dns record type,
		type=None,
		# the dns record host,
		host=None,
		# the dns record value/address,
		value=None,
		# the get_dns.records dictionary (optionally to increase speed).
		records=None,
	):
		
		# set default domain.
		if domain == None: domain = self.domain

		# get tag.
		tag = self.tag_dns(
			# the dns record type,
			type=type,
			# the dns record host,
			host=host,
			# the dns record value/address,
			value=value,)["tag"]

		# get dns.
		if records == None:
			response = self.get_dns(domain=domain)
			if response.error != None: return response
			records = response["records"]

		# handlers.
		exists = tag in list(records.keys())
		return r3sponse.success(f"Successfully checked dns record [{tag}] of namecheap user [{self.username}].", {
			"exists":exists,
			"tag":tag,
		})

		#
	def set_dns(self,
		# the domain (optional).
		domain=None,
		# the dns records (erases all others).
		records={
			"$record-1":{
				# the dns record type (required),
				"type":None,
				# the dns record host (required),
				"host":None,
				# the dns record value/address (required),
				"value":None,
				# the dns record ttl (optional default is 1800),
				"ttl":1800,
			},
		},
	):
		
		# set default domain.
		if domain == None: domain = self.domain

		# check domain.
		response = self.check_domain(domain)
		if response.error != None: return response
		elif not response["exists"]:
			return r3sponse.error(f"Namecheap user [{self.username}] does not own domain [{domain}].")

		# sld & tld.
		response = self.get_sld_and_tld(domain)
		if response.error != None: raise ValueError(response['error'])
		sld, tld = response["sld"], response["tld"]

		# api request.
		data, c = {
			"SLD":sld, 
			"TLD":tld,
		}, 0
		for tag,info in records.items():
			try:
				data[f"HostName{c}"] = info["host"]
				data[f"RecordType{c}"] = info["type"]
				data[f"Address{c}"] = info["value"]
				try:
					data[f"TTL{c}"] = info["ttl"]
				except KeyError:
					data[f"TTL{c}"] = 1800
			except KeyError:
				return r3sponse.error(f"Invalid usage, each record dictionary must have the following keys: [host, type, value].")
			c += 1
		response = self.__request__("namecheap.domains.dns.setHosts", data)
		if response["api_response"]["errors"] != None: 
			try: error = response["api_response"]["errors"]["error"]["#text"]
			except KeyError: error = response["api_response"]["errors"]
			return r3sponse.error(f"Failed to set the dns records for domain [{domain}] from namecheap user [{self.username}], error: {error}")

		# handle response.
		try:
			success = response["api_response"]["command_response"]["domain_dnsset_hosts_result"]["is_success"] == True
		except KeyError:
			success = False

		# handlers.
		if success:
			return r3sponse.success(f"Successfully set dns records for domain [{domain}] from namecheap user [{self.username}].")
		else:
			return r3sponse.error(f"Failed to set the dns records for domain [{domain}] from namecheap user [{self.username}].")

		#
	def add_dns(self,
		# the domain (optional).
		domain=None,
		# the dns record type,
		type=None,
		# the dns record host,
		host=None,
		# the dns record value/address,
		value=None,
		# the dns record ttl (optional default is 1800),
		ttl=1800,
		# the get_dns.records dictionary (optionally to increase speed).
		records=None,
	):

		# get dns records.
		if records == None:
			response = self.get_dns(domain)
			if response.error != None: return response
			records = response["records"]

		# check heroku dns records.
		response = self.check_dns(
			domain=domain,
			type=type,
			host=host,
			value=value,
			records=records,)

		# handle error.
		if response.error != None:  return response

		# handle exact duplicate.
		elif response["exists"]:
			tag = response["tag"]
			return r3sponse.error(f"DNS record [{tag}] from namecheap domain [{domain}] already exists.")

		# add new.
		else:
			
			# add dns records.
			tag = response["tag"]
			records[tag] = {
				"type":type,
				"host":host,
				"value":value,
				"ttl":ttl,
			}
			print(f"Adding DNS record [{type} {host} {value}] ...")
			response = self.set_dns(
				domain=domain,
				records=records,)
			if response.error != None: return response

			# success.
			return r3sponse.success(f"Successfully added dns record [{tag}] to namecheap domain [{domain}].", {"tag":tag})

		#
	def tag_dns(self,
		# the dns record type,
		type=None,
		# the dns record host,
		host=None,
		# the dns record value/address,
		value=None,
	):
		tag = f"{type}|{host}|{value}"
		return r3sponse.success(f"Successfully taggeed dns record [{tag}].", {"tag":tag})
	def get_sld_and_tld(self, domain=None):

		# set default domain.
		if domain == None: domain = self.domain

		# get sld & tld.
		sld, tld = "", ""
		b = domain.split('.')
		c, m, = 0, len(b)
		for i in b:
			if c < m-1:
				if sld == "":
					sld = str(i)
				else: 
					sld += '.'+i
			else:
				tld = str(i)
			c += 1

		# handlers.
		return r3sponse.success(f"Successfully retrieded the SLD & TLD of domain [{domain}].", {
			"tld":tld,
			"sld":sld,
		})
		
		#
	def get_tls(self):

		# api request.
		response = self.__request__("namecheap.ssl.getList", {})
		if response["api_response"]["errors"] != None: 
			try: error = response["api_response"]["errors"]["error"]["#text"]
			except KeyError: error = response["api_response"]["errors"]
			return r3sponse.error(f"Failed to retrieve the tls list from namecheap user [{self.username}], error: {error}")

		# handle response.
		try:
			data = response["api_response"]["command_response"]["ssllist_result"]
		except KeyError:
			return r3sponse.error(f"Failed to retrieve the tls list from namecheap user [{self.username}].")

		# iterate data.
		tls = {}
		if data == None:
			a=1 # keep empty.
		elif isinstance(data["ssl"], dict):
			tls[data["ssl"]["certificate_id"]] = data["ssl"]
		elif isinstance(data["ssl"], list):
			for _data_ in data["ssl"]:
				_data_ = self.__serialize_dictionary__(_data_)
				tls[_data_["certificate_id"]] = _data_

		# handlers.
		return r3sponse.success(f"Successfully retrieved the tls list for namecheap user [{self.username}].", {
			"tls":tls,
			"certificates":len(tls),
		})

		#
	def create_tls(self,
		# the expiration years.
		years=2,
		# the tls type.
		type="PositiveSSL",
	):

		# api request.
		response = self.__request__("namecheap.ssl.create", {
			"Years":years,
			"Type":type,
		})
		if response["api_response"]["errors"] != None: 
			try: error = response["api_response"]["errors"]["error"]["#text"]
			except KeyError: error = response["api_response"]["errors"]
			return r3sponse.error(f"Failed to create tls certificate from namecheap user [{self.username}], error: {error}")

		# handle response.
		try:
			data = response["api_response"]["command_response"]["sslcreate_result"]
		except KeyError:
			return r3sponse.error(f"Failed to create a tls certificate for namecheap user [{self.username}].")

		# handlers.
		if data["is_success"]:
			return r3sponse.success(f"Successfully created a tls certificate for namecheap user [{self.username}].", {
				"certificate_id":data["sslcertificate"]["certificate_id"],
				"order_id":data["order_id"],
				"transaction_id":data["transaction_id"],
				"charged_amount":data["charged_amount"],
				"status":data["sslcertificate"]["status"],
				"type":data["sslcertificate"]["ssltype"],
				"years":data["sslcertificate"]["years"],
			})
		else:
			return r3sponse.error(f"Failed to create a tls certificate for namecheap user [{self.username}].")

		#
	def activate_tls(self,
		# the certificate's id.
		certificate_id=None,
	):

		# check csr.
		csr = f"{self.root}/.secrets/tls/server.csr"
		if not Files.exists(csr):
			return r3sponse.error(f"There is no tls certificate present.")
		csr = Files.load(csr)

		# api request.
		response = self.__request__("namecheap.ssl.activate", {
			"CertificateID":certificate_id,
			"AdminEmailAddress":self.email,
			"CSR":csr,
			"DNSDCValidation":"true",
		})
		if response["api_response"]["errors"] != None: 
			try: error = response["api_response"]["errors"]["error"]["#text"]
			except KeyError: error = response["api_response"]["errors"]
			return r3sponse.error(f"Failed to activate tls certificate [{certificate_id}] from namecheap user [{self.username}], error: {error}")

		# handle response.
		try:
			data = response["api_response"]["command_response"]["sslactivate_result"]
		except KeyError:
			return r3sponse.error(f"Failed to retrieve the tls list from namecheap user [{self.username}].")

		# handlers.
		if not data["is_success"]:
			return r3sponse.error(f"Failed to activate tls certificate [{certificate_id}] from namecheap user [{self.username}].")

		# add dns validation record. 
		dns_validation = data["dnsdcvalidation"]["dns"]
		response = self.add_dns(
			# the domain (optional).
			domain=self.post_domain,
			# the dns record type,
			type="CNAME",
			# the dns record host,
			host=dns_validation["host_name"].split(".")[0]+self.pre_domain,
			# the dns record value/address,
			value=dns_validation["target"],
			# the get_dns.dns dictionary (optionally to increase speed).
			records=None,)
		if response.error != None and "] already exists." not in response.error: return response
		
		# handlers.
		return r3sponse.success(f"Successfully activated tls certificate [{certificate_id}] from namecheap user [{self.username}].")

		#
	# system functions.
	def __request__(self, command="namecheap.domains.getList", data={}):
		if data == {}:
			url = f"{self.base_url}&command={command}"
		else:
			url = f"{self.base_url}&command={command}&{urllib.parse.urlencode(data)}"
		response = requests.get(url)
		return self.__serialize_dictionary__(utils.__xml_to_json__(response.text))
	def __serialize_string__(self, string, banned_characters=["@"]):
		c, s, l = 0, "", False
		for char in string:
			if char not in banned_characters:
				# regular letter.
				if char.lower() == char:
					s += char.lower()
					l = False
				# capital letter.
				else:
					if c == 0:
						s += char.lower()
					else:
						if l:
							s += char.lower()
						else:
							s += "_"+char.lower()
					l = True
				c += 1
		return s
	def __serialize_dictionary__(self, response):
		_response_ = {}
		for key,value in response.items():
			s_key = self.__serialize_string__(key)
			if isinstance(value, dict):
				_response_[s_key] = self.__serialize_dictionary__(value)
			elif isinstance(value, str):
				try: integer = int(value)
				except: integer = False
				if integer != False:
					_response_[s_key] = integer
				elif value in ["false", "False", "FALSE", "DISABLED"]:
					_response_[s_key] = False
				elif value in ["true", "True", "TRUE", "ENABLED"]:
					_response_[s_key] = True
				else:
					_response_[s_key] = value
			else:
				_response_[s_key] = value
		return _response_






#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils
from w3bsite.classes import defaults as _defaults_
try:
	from django.contrib.auth.models import User as DjangoUser
	from django.contrib.auth import authenticate, login
	from django.contrib.auth import login as _login_
except Exception as e:
	if syst3m.env.get_boolean("DJANGO_RUNNING") == True:
		raise ValueError(e)

# the django object class.
class Django(_defaults_.Defaults):
	def __init__(self, 
		# the security object.
		security=None,
		# defaults.
		defaults=None,
	):

		# defaults.
		_defaults_.Defaults.__init__(self, traceback="w3bsite.Website.django",)
		self.assign(defaults.dict())

		# arguments.
		# ...

		# objects.
		self.security = security
		self.users = Users(defaults=defaults)

		#
	def start(self):

		# check migrations.
		migrations = ""
		if not Files.exists(f"{self.database_path}/data/db.sqlite3"):
			migrations = "export MIGRATIONS=True && ./manage.py migrate"

		# start django.
		print(f"Starting website [{self.root}].")
		file = File('/tmp/django.start.sh')
		if OS in ["macos"]:
			file.save(f"""
				cd {self.root}/
				source .secrets/env.sh
				export PRODUCTION=False
				export MAINTENANCE=False
				{migrations}
				./manage.py runserver
			""")
		else:
			file.save(f"""
				cd {self.root}/
				. .secrets/env.sh
				export PRODUCTION=False
				export MAINTENANCE=False
				{migrations}
				./manage.py runserver
			""")
		file.file_path.permission.set(755)
		self.django_process = subprocess.Popen(["sh", file.file_path.path])
		try:
			self.django_process.wait()
		except KeyboardInterrupt:
			self.django_process.terminate()
			os.system(f"pkill -9 -f {file.file_path.name()}")
			file.file_path.delete(forced=True)

		# handlers.
		return r3sponse.success(f"Successfully stopped website [{self.root}].")

		#
	def create(self):

		# create.
		output = utils.__execute_script__(f"""
			cd {self.root}
			django-admin startproject website
			mv website/manage.py manage.py
			mv website/website website-
			rm -fr website
			mv website- website
		""")

		# secrets.
		path = f"{self.root}/website/settings.py"
		if not Files.exists(path):
			return r3sponse.error(f"Failed to create django project [{self.root}].")
		data = Files.load(path)
		secret_key = data.split("SECRET_KEY = '")[1].split("'")[0]
		if not Files.exists(f"{self.root}/.secrets"):
			os.mkdir(f"{self.root}/.secrets")
		#Files.save(f"{self.root}/.secrets/env.sh", f"# Django environment variables.\nexport DJANGO_SECRET_KEY='{secret_key}'\n")
		#utils.__save_json__(f"{self.root}/.secrets/env.json", {f"DJANGO_SECRET_KEY":secret_key})
		self.security.set_secret_env("DJANGO_SECRET_KEY", secret_key)

		# copy urls.py
		path = f"{self.root}/website/urls.py"
		os.system(f"rm -fr {path}")
		os.system(f"cp {SOURCE_PATH}/example/website/urls.py {path}")

		# copy settings.py
		path = f"{self.root}/website/settings.py"
		os.system(f"rm -fr {path}")
		os.system(f"cp {SOURCE_PATH}/example/website/settings.py {path}")

		# copy requirements
		path = f"{self.root}/requirements/"
		os.system(f"rm -fr {path}")
		os.system(f"cp -r {SOURCE_PATH}/example/requirements {path}")

		# create requirements.pip
		#path = f"{self.root}/requirements.pip"
		#os.system(f"cp {SOURCE_PATH}/example/requirements.pip {path}")

		# copy classes
		path = f"{self.root}/classes/"
		os.system(f"rm -fr {path}")
		os.system(f"cp -r {SOURCE_PATH}/example/classes {path}")

		# copy apps
		path = f"{self.root}/apps/"
		os.system(f"rm -fr {path}")
		os.system(f"cp -r {SOURCE_PATH}/example/apps {path}")

		# copy static
		path = f"{self.root}/static/"
		os.system(f"rm -fr {path}")
		os.system(f"cp -r {SOURCE_PATH}/example/static {path}")

		# copy templates
		path = f"{self.root}/templates/"
		os.system(f"rm -fr {path}")
		os.system(f"cp -r {SOURCE_PATH}/example/templates {path}")

		# copy .gitignore
		path = f"{self.root}/.gitignore"
		os.system(f"cp {SOURCE_PATH}/example/.gitignore {path}")

		# create proc file
		path = f"{self.root}/Procfile"
		Files.save(path, "web: gunicorn website.wsgi")
		os.system(f"chmod +x {path}")

		# replace manage.py
		path = f"{self.root}/manage.py"
		data = Files.load(path)
		Files.save(path, data.replace("#!/usr/bin/env python", "#!/usr/bin/env python3"))

		# handlers.
		return r3sponse.success(f"Successfully created django website [{self.root}].")
	def create_app(self, name="home"):

		# checks.
		path = f"{self.root}/apps/{name}"
		if Files.exists(path):
			return r3sponse.error(f"App [{name}] already exists.")

		# copy.
		os.system(f"cp -r {SOURCE_PATH}/example/app {path}")

		# handlers.
		return r3sponse.success(f"Successfully created app [{name}].")

# the django database users.
class Users(_defaults_.Defaults):
	def __init__(self, 
		# defaults.
		defaults=None,
	):

		# defaults.
		_defaults_.Defaults.__init__(self, traceback="w3bsite.Website.django.users")
		self.assign(defaults.dict())

		#
	def create(self,
		username=None,
		email=None,
		password=None,
		name=None,
		superuser=False,
	):
		# check arguments.
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="create"),
			parameters={
				"username":username,
				"email":email,
				"password":password,
			})
		if not response.success: return response

		# create.
		#try:
		print("syst3m.defaults.vars.userNAME:",username)
		print("EMAIL:",email)
		print("PASSWORD:",password)

		if True:
			django_user = DjangoUser.objects.create_user(username, email, password)
			if isinstance(name, str): django_user.last_name = name
			if isinstance(superuser, bool): django_user.is_superuser = superuser
			django_user.save()

			# success.
			return r3sponse.success(f"Successfully created django user {username}.", {
				"user":django_user,
			})

		# error.
		#except Exception as e:  return r3sponse.error(f"Failed to create django user {username}, error: {e}.")

		#
	def update(self,
		# required.
		username=None,
		# optionals.
		email=None,
		password=None,
		name=None,
		superuser=None,
	):
		# check arguments.
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="update"),
			parameters={
				"username":username,
				"email":email,
				"password":password,
			})
		if not response.success: return response

		# create.
		response = self.get(username=username)
		if not response.success: return response
		django_user = response.user
		try:
			edits = 0
			if name != None: 
				edits += 1
				django_user.last_name = name
			if email != None: 
				edits += 1
				django_user.email = email
			if isinstance(superuser, bool): 
				edits += 1
				django_user.is_superuser = superuser
			if password != None: 
				edits += 1
				django_user.set_password(password)
			if edits > 0: django_user.save()

			# success.
			return r3sponse.success(f"Successfully updated django user {username}.", {
				#"user":django_user,
			})

		# error.
		except Exception as e:  return r3sponse.error(f"Failed to update django user {username}, error: {e}.")

		#
	def authenticate(self, 
		# the login credentials.
		username=None,
		password=None,
		# the request.
		request=None,
		# login the user.
		login=True,
	):

		# check arguments.
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="authenticate"),
			parameters={
				"username":username,
				"password":password,
			})
		if not response.success: return response

		# authenticate.
		try:
			user = authenticate(username=username, password=password)
			if user is not None: a=1
			else:
				response["error"] = f"Invalid verification."
				utils.__log__(f"Invalid verification attempt for user [{username}] ip address [{ip}].")
				return response

			# success.
			if login:
				_login_(request, user)

			# success.
			return r3sponse.success(f"Successfully authenticated user {username}.", {
				#"user":django_user,
			})
		# error.
		except Exception as e:  return r3sponse.error(f"Failed to login user {username}, error: {e}.")

		#
	def delete(self, username=None):

		# check arguments.
		response = r3sponse.check_parameters(
			traceback=self.__traceback__(function="delete"),
			parameters={
				"username":username,
			})
		if not response.success: return response

		# create.
		try:
			django_user = DjangoUser.objects.get(username=username)
			django_user.delete()

			# success.
			return r3sponse.success(f"Successfully deleted django user {username}.", {
				#"user":django_user,
			})

		# error.
		except Exception as e:  return r3sponse.error(f"Failed to delete django user {username}, error: {e}.")

		#
	def get(self, 
		# select one of the following user id options:
		username=None, 
		email=None,
	):

		# by username.
		identifier = None
		if username != None:
			identifier = username

			# create.
			try:
				django_user = DjangoUser.objects.get(username=username)

			# error.
			except Exception as e:  return r3sponse.error(f"Failed to retrieve django user {username}, error: {e}.")

		# by email.
		elif email != None:
			identifier = email

			# create.
			try:
				django_user = DjangoUser.objects.get(email=email)

			# error.
			except Exception as e:  return r3sponse.error(f"Failed to retrieve django user {email}, error: {e}.")

		# invalid.
		else: return r3sponse.error(f"Define one of the following parameters: [username, email].")

		# success.
		return r3sponse.success(f"Successfully retrieved django user {identifier}.", {
			"user":django_user,
			"username":django_user.username,
			"email":django_user.email,
			"name":django_user.last_name,
			"superuser":django_user.is_superuser,
		})

		#
	def exists(self, 
		# option 1:
		# by username (much faster).
		username=None,
		# option 2:
		# by email.
		email=None,
	):

		# by username.
		identifier = None
		if username != None:
			identifier = username

			# create.
			try:
				exists = DjangoUser.objects.filter(username=username).exists()

				# success.
				return r3sponse.success(f"Successfully checked the existsance of django user {username}.", {
					"exists":exists,
				})

			# error.
			except Exception as e:  return r3sponse.error(f"Failed to check the existsance of django user {username}, error: {e}.")

		# by email.
		elif email != None:
			identifier = email

			# get.
			exists = True
			response = self.get(email=email)
			if response.error != None and "User matching query does not exist" not in response.error: return response
			elif response.error != None and "User matching query does not exist" in response.error: exists = False
			return r3sponse.success(f"Successfully checked the existsance of django user {identifier}.", {
				"exists":exists,
			})

		# invalid.
		else: return r3sponse.error(f"Define one of the following parameters: [username, email].")


		#






#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils
from w3bsite.classes import defaults as _defaults_

# the security class.
class Security(_defaults_.Defaults):
	def __init__(self,
		# optional if defaults not initialized.
		root=None,
		# defaults (optional).
		defaults=None,
	):	
		
		# defaults.
		if defaults != None:
			_defaults_.Defaults.__init__(self, traceback="w3bsite.Website.security",)
			self.assign(defaults.dict())
		else:
			self.root = root

		# arguments.
		# ...

		#
	def generate_tls(self):
		# https://devcenter.heroku.com/articles/acquiring-an-ssl-certificate

		# check base.
		base = f"{self.root}/.secrets/"
		if not Files.exists(base): os.mkdir(base)
		base = f"{self.root}/.secrets/tls"
		if not Files.exists(base): os.mkdir(base)

		# check duplicate.
		if Files.exists(f"{base}/tls.key") or Files.exists(f"{base}/server.crt"):
			return r3sponse.error("The tls certificate already exists.")

		# generate.
		print("Generating a tls certificate ...")
		output = utils.__execute_script__(f"""
			cd {base}

			# Generate a passphrase
			openssl rand -base64 48 > passphrase.txt

			# Generate a Private Key
			openssl genrsa -aes128 -passout file:passphrase.txt -out server.key 4096 # 2048

			# Generate a CSR (Certificate Signing Request)
			openssl req -new -passin file:passphrase.txt -key server.key -out server.csr \
			    -subj "/C={self.country_code}/ST={self.province}/L={self.city}/O={self.organization}/OU={self.organization_unit}/CN={self.domain}/emailAddress={self.email}"

			# Remove Passphrase from Key
			cp server.key server.pass.key
			openssl rsa -in server.pass.key -passin file:passphrase.txt -out server.key

			# Do not self sign.
			# Generating a Self-Signed Certificate for 100 years
			openssl x509 -req -sha256 -days 36500 -in server.csr -signkey server.key -out server.crt

		""")

		# handler.
		if not Files.exists(f"{self.root}/.secrets/tls/server.key") or not Files.exists(f"{self.root}/.secrets/tls/server.crt"):
			os.system(f"rm -fr {base}")
			return r3sponse.error(f"Failed to generate a tls certificate.")
		else:
			return r3sponse.success(f"Successfully generated a tls certificate.")

		#
	def set_secret_env(self, key, value):
		# env.json
		if not Files.exists(f"{self.root}/.secrets"): os.mkdir(f"{self.root}/.secrets")
		try:
			env = utils.__load_json__(f"{self.root}/.secrets/env.json")
		except FileNotFoundError:
			env = {}
		splitted = key.split(".") ; c, m = 0, len(splitted)
		if m == 1:
			env[key] = value
		else:
			previous = ["Start with null at 0 index to keep c-1."]
			for s in splitted:
				c += 1
				if c == 1:
					try:
						if not isinstance(env[s], dict):
							raise ValueError(f"The specified key [{key}] can not be stored since there is already a different variable in the env.json with key [{s}] that is not a dictionary.")
					except KeyError:
						env[s] = {}
				elif c == 2:
					if c == m:	env[previous[c-1]][s] = value
					else:		
						try:
							if not isinstance(env[previous[c-1]][s], dict):
								raise ValueError(f"The specified key [{key}] can not be stored since there is already a different variable in the env.json with key [{s}] that is not a dictionary.")
						except KeyError:
							env[previous[c-1]][s] = {}
				elif c == 3:
					if c == m:	env[previous[c-2]][previous[c-1]][s] = value
					else:		
						try:
							if not isinstance(env[previous[c-2]][previous[c-1]][s], dict):
								raise ValueError(f"The specified key [{key}] can not be stored since there is already a different variable in the env.json with key [{s}] that is not a dictionary.")
						except KeyError:
							env[previous[c-2]][previous[c-1]][s] = {}
				elif c == 4:
					if c == m:	env[previous[c-3]][previous[c-2]][previous[c-1]][s] = value
					else:		
						try:
							if not isinstance(env[previous[c-3]][previous[c-2]][previous[c-1]][s], dict):
								raise ValueError(f"The specified key [{key}] can not be stored since there is already a different variable in the env.json with key [{s}] that is not a dictionary.")
						except KeyError:
							env[previous[c-3]][previous[c-2]][previous[c-1]][s] = {}
				elif c == 5:
					if c == m:	env[previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = value
					else:		
						try:
							if not isinstance(env[previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s], dict):
								raise ValueError(f"The specified key [{key}] can not be stored since there is already a different variable in the env.json with key [{s}] that is not a dictionary.")
						except KeyError:
							env[previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = {}
				elif c == 6:
					if c == m:	env[previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = value
					else:		
						try:
							if not isinstance(env[previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s], dict):
								raise ValueError(f"The specified key [{key}] can not be stored since there is already a different variable in the env.json with key [{s}] that is not a dictionary.")
						except KeyError:
							env[previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = {}
				elif c == 7:
					if c == m:	env[previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = value
					else:		
						try:
							if not isinstance(env[previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s], dict):
								raise ValueError(f"The specified key [{key}] can not be stored since there is already a different variable in the env.json with key [{s}] that is not a dictionary.")
						except KeyError:
							env[previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = {}
				elif c == 8:
					if c == m:	env[previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = value
					else:		
						try:
							if not isinstance(env[previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s], dict):
								raise ValueError(f"The specified key [{key}] can not be stored since there is already a different variable in the env.json with key [{s}] that is not a dictionary.")
						except KeyError:
							env[previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = {}
				elif c == 9:
					if c == m:	env[previous[c-8]][previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = value
					else:		
						try:
							if not isinstance(env[previous[c-8]][previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s], dict):
								raise ValueError(f"The specified key [{key}] can not be stored since there is already a different variable in the env.json with key [{s}] that is not a dictionary.")
						except KeyError:
							env[previous[c-8]][previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = {}
				elif c == 10:
					if c == m:	env[previous[c-9]][previous[c-8]][previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = value
					else:		
						try:
							if not isinstance(env[previous[c-9]][previous[c-8]][previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s], dict):
								raise ValueError(f"The specified key [{key}] can not be stored since there is already a different variable in the env.json with key [{s}] that is not a dictionary.")
						except KeyError:
							env[previous[c-9]][previous[c-8]][previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = {}
				elif c == 11:
					if c == m:	env[previous[c-10]][previous[c-9]][previous[c-8]][previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = value
					else:		
						try:
							if not isinstance(env[previous[c-10]][previous[c-9]][previous[c-8]][previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s], dict):
								raise ValueError(f"The specified key [{key}] can not be stored since there is already a different variable in the env.json with key [{s}] that is not a dictionary.")
						except KeyError:
							env[previous[c-10]][previous[c-9]][previous[c-8]][previous[c-7]][previous[c-6]][previous[c-5]][previous[c-4]][previous[c-3]][previous[c-2]][previous[c-1]][s] = {}
				else:
					raise ValueError(f"Reached max recursive depth with key [{key}].")
				previous.append(s)
		utils.__save_json__(f"{self.root}/.secrets/env.json", env)
		# env.sh
		try:
			env = Files.load(f"{self.root}/.secrets/env.sh")
		except FileNotFoundError:
			env = ""
		l_key = key.replace(".","-")
		replaced_value = value.replace("\n", "\\n")
		if f'export {l_key}=' in env:
			_env_ = ""
			for line in env.split("\n"):
				if f'export {l_key}=' in line:
					_env_ += f'\nexport {l_key}="{replaced_value}"'
				else:
					_env_ += line+"\n"
		else:
			env += f'\nexport {l_key}="{replaced_value}"'
		for i in range(100):
			if "\n\n" in env: env = env.replace("\n\n","\n")
			else: break
		Files.save(f"{self.root}/.secrets/env.sh", env)
		return r3sponse.success(f"Successfully setted the secret environment variable [{key}].")
	def get_secret_env(self, key, default=None):
		value = syst3m.env.get_string(key, default=default)
		if value == "None": value = None
		if value == None: raise ValueError(f"Secret enironment variable [{key}] is undefined.")
		if "\\n" in value: value = value.replace("\\n", "\n")
		return value


#!/usr/bin/env python3

# Note!
# ' are required, do not use any '.

# setup.
from setuptools import setup, find_packages
setup(
	name='w3bsite',
	version='4.7.8',
	description='Some description.',
	url='http://github.com/vandenberghinc/w3bsite',
	author='Daan van den Bergh',
	author_email='vandenberghinc.contact@gmail.com',
	license='MIT',
	packages=find_packages(),
	zip_safe=False,
	include_package_data=True,
	install_requires=[
        #'ssht00ls',
        #'fil3s',
        #'encrypti0n',
        #'syst3m',
        #'netw0rk',
        #'cl1',
        #'r3sponse',
        'django',
        'xmltodict',
        #'firebase_admin',
        'stripe',
        'requests',
    ])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *

# the database class.
class Database(syst3m.objects.Traceback):
	def __init__(self, firestore=None, path=None):
		
		# traceback.
		syst3m.objects.Traceback.__init__(self, traceback="w3bsite.Website.database")

		# attributes.
		self.firestore = firestore
		self.path = gfp.clean(path)
		self.file_path = self.fp = FilePath(self.path)
		if not self.fp.exists():
			r3sponse.log(f"&ORANGE&Root permission&END& required to create database [{self.path}].")
			os.system(f"sudo mkdir {self.path}")
			Files.chmod(path=self.path, permission=700, sudo=True)
			Files.chown(path=self.path, owner=syst3m.defaults.vars.user, group=syst3m.defaults.vars.group, sudo=True)
		self.cache = None
		if self.firestore != None:
			self.cache = syst3m.cache.Cache(path=self.path)

		#
	def load(self, path=None, format="str"):
		if path == None: return r3sponse.error(self.__traceback__(function="load")+" Define parameter: [path].")
		if self.firestore != None:
			return self.firestore.load(path)
		elif self.cache != None:
			return self.cache.load(group=path, format=format)
	def save(self, path=None, data=None, format="str"):
		if path == None: return r3sponse.error(self.__traceback__(function="save")+" Define parameter: [path].")
		if self.firestore != None:
			return self.firestore.save(path, data)
		elif self.cache != None:
			return self.cache.save(group=path, data=data, format=format)
	def delete(self, path=None, data=None, format="str"):
		if path == None: return r3sponse.error(self.__traceback__(function="delete")+" Define parameter: [path].")
		if self.firestore != None:
			return self.firestore.delete(path)
		elif self.cache != None:
			return self.cache.delete(group=path)
	# str representation.
	def __str__(self):
		return str(self.fp.path)
	
			